/*
 * Copyright (c) 1997 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

/*
 * This file contains a set of tests using the functions defined in
 * mig-tests.defs.  
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <mach/message.h>
#include <mach/port.h>
#include <mach/mig_errors.h>
#include <memory.h>

#include "mig-tests-client.h"

/* macro for calling a client stub */
#define _try(_func, _params) {\
	int _result;\
	printf("Calling function %s...\n", #_func);\
	if ((_result = _func _params) != KERN_SUCCESS) {\
		fprintf(stderr, "%s%s failed, result = %d (0x%08x)\n",\
			#_func, #_params, _result, _result);\
	}\
}

/* function to print an array of integers */
void parr(int *a, int size)
{
	int i;
	printf("{ ");
	for (i=0; i<size; i++) {
		if (i) printf(", ");
		printf("%d",a[i]);
	}
	printf(" }\n");
}

/* print out a port type */
char * typestr(mach_port_t p)
{
	static char str[500];
	mach_port_type_t tp;
	int ret;
	
	ret = mach_port_type(mach_task_self(), p, &tp);
	if (ret == KERN_INVALID_NAME) return "INVALID_NAME";
	if (ret != KERN_SUCCESS) return "UNKNOWN";
	str[0] = 0;
	if (tp & MACH_PORT_TYPE_SEND) strcat(str, "SEND ");
	if (tp & MACH_PORT_TYPE_SEND_ONCE) strcat(str, "SEND_ONCE ");
	if (tp & MACH_PORT_TYPE_RECEIVE) strcat(str, "RECV ");
	if (tp & MACH_PORT_TYPE_PORT_SET) strcat(str, "PORT_SET ");
	if (tp & MACH_PORT_TYPE_DEAD_NAME) strcat(str, "DEAD ");
	if (tp & MACH_PORT_TYPE_DNREQUEST) strcat(str, "DNREQ ");
	return str;
}

/*
 * This is the main body of the client-side tests.
 */
int mig_tests_client(mach_port_t right)
{
	int i;

	/*
	 * IN-LINE TESTS
	 */

	/* integers */
	{
		int a = 20;
		int b = 10;
		int c = 0;
		printf("\n\nBefore call, a=%d, b=%d, c=%d\n", a, b, c);
		_try(inline_integers, (right, a, &b, &c));
		printf("After call, a=%d, b=%d, c=%d\n", a, b, c);
	}
	/* fixed structs */
	{
		IFstruct a, b, c;
		memset(&a, 1, sizeof(a));
		memset(&b, 2, sizeof(b));
		memset(&c, 0, sizeof(c));
		printf("\n\nBefore call, a=");parr((int *)&a, 100);
		printf("b=");parr((int *)&b, 100);
		printf("c=");parr((int *)&c, 100);
		_try(inline_fixed_structs, (right, a, &b, &c));
		printf("After call, a=");parr((int *)&a, 100);
		printf("b=");parr((int *)&b, 100);
		printf("c=");parr((int *)&c, 100);
	}
	/* fixed arrays */
	{
		IFarray a, b, c;
		for (i=0; i<300; i++) {
			a[i] = i;
			b[i] = i;
			c[i] = 0;
		}
		printf("\n\nBefore call, a=");parr(a, 300);
		printf("b=");parr(b, 300);
		printf("c=");parr(c, 300);
		_try(inline_fixed_arrays, (right, a, b, c));
		printf("After call, a=");parr(a, 300);
		printf("b=");parr(b, 300);
		printf("c=");parr(c, 300);
	}
	/* bounded arrays */
	{
		IBarray a, b, c;
		for (i=0; i<100; i++) {
			a[i] = i;
			b[i] = i;
			c[i] = 0;
		}
		printf("\n\nBefore call, a=");parr(a, 10);
		/*printf("b=");parr(b, 10);*/
		printf("c=");parr(c, 10);
		_try(inline_bounded_arrays, (right, a, 10, &c, &i));
		printf("After call, i=%d, a=",i);parr(a, 10);
		/*printf("b=");parr(b, 10);*/
		printf("c=");parr(c, i);
		
		for (i=0; i<500; i++) {
			a[i] = i;
			b[i] = i;
			c[i] = 0;
		}
		printf("\n\nBefore call, a=");parr(a, 500);
		/*printf("b=");parr(b, 500);*/
		printf("c=");parr(c, 500);
		_try(inline_bounded_arrays, (right, a, 500, &c, &i));
		printf("After call, i=%d, a=", i);parr(a, 500);
		/*printf("b=");parr(b, 500);*/
		printf("c=");parr(c, i);
	}
	/* unbounded arrays */
	{
		/* NOTE:
		   Currently Flick allocates all out parameters,
		   regardless of whether a buffer was passed in or
		   not.  If a buffer is allocated and passed in, it
		   would NOT be de/reallocated.  This differs slightly
		   from MIG; however, assuming one always passes in a
		   null pointer and size 0, both MIG and Flick produce
		   the same effect. */
		
		Iarray a, c;
		int ac = 100, cc = 0;
		a = (unsigned *)malloc(sizeof(unsigned)*ac);
		c = 0;
		for (i=0; i<100; i++) {
			a[i] = i;
		}
		printf("\n\nBefore call, a[%d]=",ac);parr(a, ac);
		/*printf("b[%d]=",bc);parr(b, bc);*/
		printf("c[%d]=",cc);parr(c, cc);
		_try(inline_unbounded_arrays, (right, a, ac, &c, &cc));
		printf("After call, a[%d]=",ac);parr(a, ac);
		/*printf("b[%d]=",bc);parr(b, bc);*/
		printf("c[%d]=",cc);parr(c, cc);
		vm_deallocate(mach_task_self(), (vm_address_t) c, cc);
	}
	/* fixed strings */
	{
		IFstring a = "This is a very profound sentence.",
			 b = "Sentence two expands on sentence one's profoundness.",
			 c = "Our third sentence is quite boring.";
		printf("\n\nBefore call, a[%ld]= %s\n",strlen(a),a);
		printf("b[%ld]= %s\n",strlen(b),b);
		printf("c[%ld]= %s\n",strlen(c),c);
		_try(inline_fixed_strings, (right, a, b, c));
		printf("After call, a[%ld]= %s\n",strlen(a),a);
		printf("b[%ld]= %s\n",strlen(b),b);
		printf("c[%ld]= %s\n",strlen(c),c);
	}
	/* bounded strings */
	{
		IFstring a = "This is a very profound sentence.",
			 b = "Sentence two expands on sentence one's profoundness.";
		printf("\n\nBefore call, a[%ld]= %s\n",strlen(a),a);
		printf("b[%ld]= %s\n",strlen(b),b);
		_try(inline_bounded_strings, (right, a, b));
		printf("After call, a[%ld]= %s\n",strlen(a),a);
		printf("b[%ld]= %s\n",strlen(b),b);
	}
	/* port rights */
	{
		mach_port_t a, b, c;
		
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &a);
		b = right;
		c = MACH_PORT_NULL;
		printf("\n\nBefore call, a = 0x%08x (%s)\n", a, typestr(a));
		printf("b = 0x%08x (%s)\n", b, typestr(b));
		printf("c = 0x%08x (%s)\n", c, typestr(c));
		_try(inline_port_rights, (right, a, &b, &c));
		printf("After call, a = 0x%08x (%s)\n", a, typestr(a));
		printf("b = 0x%08x (%s)\n", b, typestr(b));
		printf("c = 0x%08x (%s)\n", c, typestr(c));
		mach_port_deallocate(mach_task_self(), c);
	}
	/* port right arrays */
	{
		Imach_port_array_t a, b;
		int ac = 2, bc = 0;
		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    sizeof(mach_port_t)*2, 1);
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &(a[0]));
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &(a[1]));
		b = MACH_PORT_NULL;
		printf("\n\nBefore call, a[0] = 0x%08x (%s),   a[1] = 0x%08x (%s)\n",
		       a[0], typestr(a[0]), a[1], typestr(a[1]));
		_try(inline_port_arrays, (right, a, ac, &b, &bc));
		printf("After call, a[0] = 0x%08x (%s),   a[1] = 0x%08x (%s)\n",
		       a[0], typestr(a[0]), a[1], typestr(a[1]));
		printf("b[%d]: b[0] = 0x%08x (%s),   b[1] = 0x%08x (%s)\n", bc,
		       b[0], typestr(b[0]), b[1], typestr(b[1]));
		mach_port_deallocate(mach_task_self(), b[0]);
		mach_port_deallocate(mach_task_self(), b[1]);
	}
	
	/*
	 * OUT-OF-LINE TESTS
	 */

	/* Only variable-length out-of-line data works under
           MIG. Fixed length data is set to be of length 0, and thus
           nothing is transmitted.  In Flick, this problem is
           resolved, thus any type of out-of-line data can be
           sent/received correctly. */

	/* integers  -- CURRENTLY DON'T WORK UNDER MIG */
	{
		Oint a, b, c = 0;
		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    sizeof(int), 1);
		*a = 10;
		vm_allocate(mach_task_self(), (vm_address_t *) &b,
			    sizeof(int), 1);
		*b = 20;
		printf("\n\nBefore call, a=0x%08x,%d, b=0x%08x,%d, c=0x%08x\n",
		       (int)a, *a, (int)b, *b, (int)c);
		_try(outline_integers, (right, a, &b, &c));
		printf("After call, a=0x%08x,%d, b=0x%08x,%d, c=0x%08x,%d\n",
		       (int)a, *a, (int)b, *b, (int)c, *c);
	}
	/* fixed structs  -- CURRENTLY DON'T WORK UNDER MIG */
	{
		OFstruct a, b, c;
		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    sizeof(*a), 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &b,
			    sizeof(*b), 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &c,
			    sizeof(*c), 1);
		memset(a, 1, sizeof(*a));
		memset(b, 2, sizeof(*b));
		memset(c, 0, sizeof(*c));
		printf("\n\nBefore call, a=");parr((int *)a, 200);
		printf("b=");parr((int *)b, 200);
		printf("c=");parr((int *)c, 200);
		_try(outline_fixed_structs, (right, a, &b, &c));
		printf("After call, a=");parr((int *)a, 200);
		printf("b=");parr((int *)b, 200);
		printf("c=");parr((int *)c, 200);
	}
	/* fixed arrays  -- CURRENTLY DON'T WORK UNDER MIG */
	{
		OFarray a, b, c;
		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    sizeof(int)*400, 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &b,
			    sizeof(int)*400, 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &c,
			    sizeof(int)*400, 1);
		for (i=0; i<400; i++) {
			(*a)[i] = i;
			(*b)[i] = i;
			(*c)[i] = 0;
		}
		printf("\n\nBefore call, a=");parr(*a, 400);
		printf("b=");parr(*b, 400);
		printf("c=");parr(*c, 400);
		_try(outline_fixed_arrays, (right, a, &b, &c));
		printf("After call, a=");parr(*a, 400);
		printf("b=");parr(*b, 400);
		printf("c=");parr(*c, 400);
	}
	/* unbounded arrays */
	{
		Oarray a, b, c;
		int ac = 100, bc = 200, cc = 0;
		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    sizeof(int)*ac, 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &b,
			    sizeof(int)*bc, 1);
		c = 0;
		for (i=0; i<ac; i++) {
			a[i] = i;
		}
		for (i=0; i<bc; i++) {
			b[i] = i * i;
		}
		printf("\n\nBefore call, a[%d]=",ac);parr(a, ac);
		printf("b[%d]=",bc);parr(b, bc);
		printf("c[%d]=",cc);parr(c, cc);
		_try(outline_unbounded_arrays, (right, a, ac, &b, &bc, &c, &cc));
		printf("After call, a[%d]=",ac);parr(a, ac);
		printf("b[%d]=",bc);parr(b, bc);
		printf("c[%d]=",cc);parr(c, cc);
		vm_deallocate(mach_task_self(), (vm_address_t) c, cc);
	}
	/* fixed strings  -- CURRENTLY DON'T WORK UNDER MIG */
	{
		OFstring a, b, c;

		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    800, 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &b,
			    800, 1);
		vm_allocate(mach_task_self(), (vm_address_t *) &c,
			    800, 1);
		strcpy(*a, "This is a very profound sentence.");
		strcpy(*b, "Sentence two expands on sentence one's profoundness.");
		strcpy(*c, "Our third sentence is quite boring.");
		printf("\n\nBefore call, a[%ld]= %s\n",strlen(*a),*a);
		printf("b[%ld]= %s\n",strlen(*b),*b);
		printf("c[%ld]= %s\n",strlen(*c),*c);
		_try(outline_fixed_strings, (right, a, &b, &c));
		printf("After call, a[%ld]= %s\n",strlen(*a),*a);
		printf("b[%ld]= %s\n",strlen(*b),*b);
		printf("c[%ld]= %s\n",strlen(*c),*c);
	}
	/* port right arrays */
	{
		mach_port_array_t a, b, c;
		int ac = 2, bc = 0, cc = 0;
		mach_port_t bs1, bs2;
		vm_allocate(mach_task_self(), (vm_address_t *) &a,
			    sizeof(mach_port_t)*2, 1);
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &(a[0]));
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &(a[1]));
		vm_allocate(mach_task_self(), (vm_address_t *) &b,
			    sizeof(mach_port_t)*2, 1);
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &(b[0]));
		mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &(b[1]));
		bs1 = b[0];
		bs2 = b[1];
		c = MACH_PORT_NULL;
		printf("\n\nBefore call, a[0] = 0x%08x (%s),   a[1] = 0x%08x (%s)\n",
		       a[0], typestr(a[0]), a[1], typestr(a[1]));
		printf("b[%d]: b[0] = 0x%08x (%s),   b[1] = 0x%08x (%s)\n", bc,
		       b[0], typestr(b[0]), b[1], typestr(b[1]));
		_try(outline_port_arrays, (right, a, ac, &b, &bc, &c, &cc));
		printf("After call, a[0] = 0x%08x (%s),   a[1] = 0x%08x (%s)\n",
		       a[0], typestr(a[0]), a[1], typestr(a[1]));
		printf("b[%d]: b[0] = 0x%08x (%s),   b[1] = 0x%08x (%s)\n", bc,
		       b[0], typestr(b[0]), b[1], typestr(b[1]));
		printf("c[%d]: c[0] = 0x%08x (%s),   c[1] = 0x%08x (%s)\n", cc,
		       c[0], typestr(c[0]), c[1], typestr(c[1]));
		mach_port_deallocate(mach_task_self(), a[0]);
		mach_port_deallocate(mach_task_self(), a[1]);
		mach_port_deallocate(mach_task_self(), bs1);
		mach_port_deallocate(mach_task_self(), bs2);
		mach_port_deallocate(mach_task_self(), c[0]);
		mach_port_deallocate(mach_task_self(), c[1]);
	}

	return 0;
}

/* This is the client function called by the test client. */
int call_client(mach_port_t right) {
	printf("About to run client...\n");
	printf("Returning from client(%d)...\n", mig_tests_client(right));
	return 0;
}

/* End of file. */
